/*
 * Copyright (C) 2007 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include<iostream>

#include "PlanTreeTaskListNode.hh"
#include "TaskListNode.hh"

using namespace std;

/**
 * constructor: makes a "copy" of TLn but with its task in a string
 * in the nodes, copy also IDs
 */
PlanTreeTaskListNode::PlanTreeTaskListNode(TaskListNode const& TLn)
	: instance(TLn.getInstancePtr()), ID(TLn.getID())
{
	// predecessors and successors will be set by PlanTreeTaskList class
}

PlanTreeTaskListNode::PlanTreeTaskListNode(std::shared_ptr<TaskInstance> instance, unsigned long int const ID, std::vector<std::string> const& involved_agents)
	:instance(instance), ID(ID), involved_agents(involved_agents)
{}

/**
 * Another functions needs to be called to restore the predecessors and successors lists
 */
PlanTreeTaskListNode* PlanTreeTaskListNode::makeCopy() const{
	return new PlanTreeTaskListNode(std::make_shared<TaskInstance>(*instance), ID, involved_agents);
}

/**
 * adds s in successors if it is not already in it
 */
void PlanTreeTaskListNode::naiveAddSuccessor(PlanTreeTaskListNode* s)
{
	if(s == NULL)
	{
		cerr << "ERROR: trying to add a not exisitng PlanTreeTaskListNode";
		cerr << " in a PlanTreeTaskListNode successors" << endl;
	}
	else if( find(predecessors.begin(), predecessors.end(), s) != predecessors.end() )
	{
		// s is already in predecessors (it will make a loop)
		cerr << "ERROR: trying to add a PlanTreeTaskListNode in a PlanTreeTaskListNode";
		cerr << " successors whereas it is already in predecessors" << endl;
	}
	else if( find(successors.begin(), successors.end(), s) == successors.end() )
	{
		// everything OK, s is not already in successors
		successors.push_back(s);
	}
}

/**
 * add p in predecessors if it is not already in it
 */
void PlanTreeTaskListNode::naiveAddPredecessor(PlanTreeTaskListNode* p)
{
	if(p == NULL)
	{
		cerr << "ERROR: trying to add a not exisitng PlanTreeTaskListNode";
		cerr << " in a PlanTreeTaskListNode predecessors" << endl;
	}
	else if( find(successors.begin(), successors.end(), p) != successors.end() )
	{
		// s is already in successors (it will make a loop)
		cerr << "ERROR: trying to add a PlanTreeTaskListNode in a PlanTreeTaskListNode";
		cerr << " predecessors whereas it is already in successors" << endl;
	}
	else if( find(predecessors.begin(), predecessors.end(), p) == predecessors.end() )
	{
		// everything OK, s is not already in predecessors
		predecessors.push_back(p);
	}
}

/**
 * returns true if PTTLn has the same successors as the
 * current PlanTreeTaskListNode (the comparison of successors
 * does not consider node IDs)
 */
bool PlanTreeTaskListNode::hasSameSuccessors(PlanTreeTaskListNode const* PTTLn) const
{
	vector<PlanTreeTaskListNode*> PTTLnSuc = PTTLn->getSuccessors();

	if(PTTLnSuc.size() == successors.size())
	{
		unsigned int i,j;
		for(i=0 ; i<successors.size() ; ++i)
		{
			for(j=0 ; j<PTTLnSuc.size() ; ++j)// we explore PTTLnSuc to find a similar successor
				if(successors[i]->getText() == PTTLnSuc[j]->getText())
					break;

			if(j == PTTLnSuc.size())
				return false;
		}

		return true;
	}

	return false;
}

/**
 * returns true if PTTLn has the same predecessors as the current PlanTreeTaskListNode
 * (the comparison of predecessors is based on string comparison and does not consider node IDs)
 */
bool PlanTreeTaskListNode::hasSamePredecessors(PlanTreeTaskListNode const* PTTLn) const
{
	vector<PlanTreeTaskListNode*> PTTLnPred = PTTLn->getPredecessors();

	if(PTTLnPred.size() == predecessors.size())
	{
		unsigned int i,j;
		for(i=0 ; i<predecessors.size() ; ++i)
		{
			for(j=0 ; j<PTTLnPred.size() ; ++j)// we explore PTTLnPred to find a similar predecessor
				if(predecessors[i]->getText() == PTTLnPred[j]->getText())
					break;

			if(j == PTTLnPred.size())
				return false;
		}

		return true;
	}

	return false;
}
