/*
 * Copyright (C) 2007 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include<iostream>
#include "ChoiceReminder.hh"

using namespace std;

/** Constructor.
 * Set rootCP to root_point and current_point
 * @param rootCP Root point for the reminder (can not be NULL)
 */
ChoiceReminder::ChoiceReminder(ChoicePoint* rootCP) : root_point(rootCP)
{
	points.insert(root_point);
	current_point = root_point;
	current_point->setTag(ChoicePoint::CURRENT_TAG);
}

/** Constructor.
 * Creates a root node from TL and sets current node to root node
 * @param wb A LightWorldBase is built from this parameter and given to the new root node
 * @param TL The TaskList considered
 * @param app The PartialPlanAppraiser for the root node
 */
ChoiceReminder::ChoiceReminder(WorldBase const& wb, TaskList const& TL, PartialPlanAppraiser const& app)
{
	PlanTree PT(TL);
	LightWorldBase lwb(wb);
	root_point = new ChoicePoint(ChoicePoint::ROOT_TYPE, lwb, TL, vector<unsigned long int>(0), PT, app);

	points.insert(root_point);
	current_point = root_point;
	current_point->setTag(ChoicePoint::CURRENT_TAG);
}

/** Copy constructor.
 * @param cr ChoiceReminder to copy
 */
ChoiceReminder::ChoiceReminder(ChoiceReminder const& cr)
{
	set<ChoicePoint*>::const_iterator it1;
	set<ChoicePoint*>::iterator it2;
	ChoicePoint* cp;
	const set<ChoicePoint*>& crPoints = cr.getChoicePoints();
	vector<ChoicePoint*> children;
	map<ChoicePoint*, ChoicePoint*> crCPandCurrentCP; //Keep track of the ChoicePoint created, translate a point from cr (to copy) to the new created points

	// we make a copy of all choice points
	for(it1 = crPoints.begin() ; it1 != crPoints.end() ; ++it1) {
		cp = new ChoicePoint(*(*it1));
		points.insert(cp);
		// we keep the correspondance between cr ChoicePoint adresses and current ChoicePoint adresses
		crCPandCurrentCP[*it1] = cp;
	}
	// one all points are created we update father and children fields for each member of points
	for(it1 = crPoints.begin(), it2 = points.begin() ; it1 != crPoints.end() && it2 != points.end() ; ++it1, ++it2) {
		// we update father of point it2
		cp = (*it1)->getFather();
		if(cp == NULL) (*it2)->setFather(NULL);
		else (*it2)->setFather(crCPandCurrentCP[cp]);

		// we update children of point it2
		children = (*it1)->getChildren();
		for(unsigned long int i=0 ; i<children.size() ; ++i) {
			cp = children[i];
			if(cp == NULL) (*it2)->addChild(NULL);
			else (*it2)->addChild(crCPandCurrentCP[cp]);
		}
	} // end update father and children of each points

	root_point = crCPandCurrentCP[cr.getRootPoint()];
	current_point = crCPandCurrentCP[cr.getCurrentPoint()];
}

/** All the elements of the vector becomes children of the current node.
 * If vec is empty do nothing
 * @param vec New children of the the current node
 */
void ChoiceReminder::addChoicesToCurrentNode(const vector<ChoicePoint*>& vec)
{
	if(vec.size()==0)
		cerr << "WARNING: trying to add nodes to the current node in the ChoiceReminder there is no node to add (operation ignored)" << endl;
	else {
		for(unsigned long int i=0 ; i<vec.size() ; ++i) {
			if(vec[i]==NULL)
				cerr << "ERROR: when adding nodes to the ChoiceReminder current node, a ChoicePoint to add has a NULL pointer (ChoicePoint not added)" << endl;
			else {
				points.insert(vec[i]);
				current_point->addChild(vec[i]);
			}
		}
	}
}

/** Backtrack in the tree of choice points.
 * First calls update(), if it returns true it returns true (meaning
 * that all the tree has been explored) else get an unexplored
 * child of current_point and sets it as the current node (and return false)
 * @return True if the whole tree has been explored, otherwise change the current point and returns false
 */
bool ChoiceReminder::backtrack()
{
	if( this->update() ) return true;

	vector<ChoicePoint*> CurrentChildren = current_point->getChildren();
	for(unsigned long int i=0 ; i<CurrentChildren.size() ; ++i)
		if(CurrentChildren[i]->getTag() == ChoicePoint::UNEXPLORED_TAG) {
			current_point = CurrentChildren[i];
			current_point->setTag(ChoicePoint::CURRENT_TAG);
			return false;
		}
	
	cerr << "ERROR: when trying to make a backtrack, the ChoicePoint reached by update() function has no UNEXPLORED children" << endl;
	return false;
}

/** Update the choice reminder by changing the current node to the first unexplored.
 * To do so calls propagateTag for the current node if tag of this node is FAILED or SUCCESS,
 * go up on the tree to reach a ChoicePoint with tag == UNEXPLORED_TAG and sets current_point to it.
 * While going up in the tree all children form node with a tag equivalent to EXPLORED (FAILED, SUCCESS, ...)
 * are removed.
 * @return True if the whole tree has been explored, false otherwise
 */
bool ChoiceReminder::update()
{
	if(current_point == NULL)
		cerr << "ERROR: trying to update the ChoiceReminder whereas its current node is NULL" << endl;
	else {
		vector<ChoicePoint*>::const_iterator it1;
		set<ChoicePoint*>::const_iterator it2;
		
		ChoicePoint::Tag result = current_point->propagateTags();
		
		while(result == ChoicePoint::FAILED_TAG || result == ChoicePoint::SUCCESS_TAG || result == ChoicePoint::EXPLORED_TAG || result == ChoicePoint::COST_EXCEEDED_TAG) {
			if(current_point->getType() == ChoicePoint::ROOT_TYPE) return true;
			
			// we get all children of the current point and we remove them since no longer needed
			// to do so first remove them from the set then call a method to delete them all at once
			// (usefull for reducing memory needs especially because STL containers do not free memory when clearing them)
			vector<ChoicePoint*> const& vec = current_point->getChildren();
			for(it1 = vec.begin() ; it1!=vec.end() ; ++it1) {
				it2 = points.find(*it1);
				if(it2 != points.end()) {
					points.erase(it2);
				}
			}
			current_point->deleteChildren();
			current_point = current_point->getFather();
			result = current_point->propagateTags();
		}
	}
	return false;
}

/** Select the first unexplored child on current_point.
 * It also sets current_point to it
 */
void ChoiceReminder::selectUnexploredChildOfCurrentPoint()
{
	unsigned long int i;
	vector<ChoicePoint*> children = current_point->getChildren();
	for(i=0 ; i<children.size() ; ++i)
		if(children[i]->getTag() == ChoicePoint::UNEXPLORED_TAG) {
			current_point = children[i];
			current_point->setTag(ChoicePoint::CURRENT_TAG);
			break;
		}
	if(i==children.size()) {
		cerr << "ERROR: when trying to select an unexplored child of a ChoicePoint, ";
		cerr << "no such child has been found (current_point maintained to previous value)" << endl;
	}
}

/** Return the list of all choice point with the same father as current_point.
 * if the current point is a parallel one, gets all IDs of the tasks
 * associated to its brothers and returns it
 * else returns an empty vector
 * @return List of point with the same father (or empty vector)
 */
vector<unsigned long int> ChoiceReminder::getCurrentFatherChoices() const
{
	if(current_point->getType() == ChoicePoint::PARALLEL_TYPE) {
		vector<ChoicePoint*> brothers = current_point->getFather()->getChildren(); // contains this choice point and its brothers
		vector<unsigned long int> ids(brothers.size());
		for(unsigned long int i=0 ; i<brothers.size() ; ++i) ids[i] = brothers[i]->getTask();
		return ids;
	}
	return vector<unsigned long int>(0);
}

/** Return a string that represents this choice reminder.
 * @return A string representing this ChoiceReminder
 */
std::string ChoiceReminder::toString() const
{
	return recursiveBranchPrint(root_point, "");
}

/** Actual implementation for the ChoiceReminder::toString().
 * Prints the point given as argument (and add the offset before printing it)
 * @param point Current point to print
 * @param offset Offset to use
 * @return The string representing the choice point in parameter
 */
std::string ChoiceReminder::recursiveBranchPrint(ChoicePoint* point, std::string offset) const
{
	std::vector<ChoicePoint*> const& children(point->getChildren());
	std::vector<ChoicePoint*>::const_iterator it;
	std::string result;
	
	result = offset + point->toString() + "\n";

	for(it=children.begin(); it!=children.end(); ++it){
		result+=recursiveBranchPrint(*it, offset+"   ");
	}

	return result;
}

/** Destructor.
 * Deletes all members of all points and empty the set of points
 */
ChoiceReminder::~ChoiceReminder()
{
	for(set<ChoicePoint*>::iterator it = points.begin() ; it != points.end() ; ++it) delete *it;
	points.clear();
}
