/*
 * Copyright (C) 2007 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#if !defined HATP_PLAN_STOCK_MANAGER_HH_
#define HATP_PLAN_STOCK_MANAGER_HH_

#include<map>
#include<vector>
#include<boost/date_time/posix_time/posix_time.hpp>

#include "Plan.hh"
#include "PlanTree.hh"
#include "TaskList.hh"

/** Handles the plan found by the core.
 */
class PlanStockManager
{
private:
	/** All the plans (the key of the map is the score of the plan) */
	std::multimap<float, Plan*> plans;

	/** Time when planning started */
	boost::posix_time::ptime start_time;

	/** Current number of plans found (whether there are kept or not) */
	unsigned long int NumberOfPlans;

	/** Whether a solution exists or not */
	bool solutionExists;

	/** Last message to send */
	std::string lastMessage;

public:
	explicit PlanStockManager();

	void setSolutionExists(bool b);
	bool getSolutionExists() const;
	float getBestScore() const;
	std::map<std::string, float> const* getBestPlanDetails() const;

	void addPlan(Plan* p);
	void setStartTime(boost::posix_time::ptime const& argTime);

	int getNumberOfPlans() const;
	std::vector<Plan*> getPlans() const;
	Plan* getBestPlan() const;
	Plan* getNextPlan();
	Plan* getPlanByID(unsigned long int id) const;
	std::vector<Plan*> getBestPlans() const;
	unsigned int getNbPlans() const;

	std::string getLastMessage() const;

	void clear();
	~PlanStockManager();
};

/** Constructor: create an empty set of plans.
 */
inline PlanStockManager::PlanStockManager() {
	NumberOfPlans = 0;
	solutionExists = false;
	lastMessage = "";
}

/** Change the fact that at least one plan exists.
 * @param b Tell whether a plan exists or not
 */
inline void PlanStockManager::setSolutionExists(bool b) {
	solutionExists = b;
}

/** Return whethe a plan exists or not.
 * @return True if a plan exists
 */
inline bool PlanStockManager::getSolutionExists() const {
	return solutionExists;
}

/** Get the last message built by the PlanStockManager.
 * The message are then sent to hatpconsole to display information about the plans
 * @return A message to display on the hatpconsole
 */
inline std::string PlanStockManager::getLastMessage() const {
	return lastMessage;
}

/** Return the bast plan score if a plan exists, -1 othewise.
 * @return -1 if plans is empty else return the score of the first member in plans
 */
inline float PlanStockManager::getBestScore() const {
	if(plans.empty()) return -1;
	else return (plans.begin()->first);
}


/** Return the number of plans found
 * @return The number of plans
 */
inline int PlanStockManager::getNumberOfPlans() const {
	return NumberOfPlans;
}

/** Return the details on the current best plan (NULL if no plan)
 * @return NULL if plans is empty else return a pointer on score details of the first member in plans
 */
inline std::map<std::string, float> const* PlanStockManager::getBestPlanDetails() const {
	if(plans.empty()) return (std::map<std::string, float> const*)NULL;
	else return plans.begin()->second->getPtrScoreDetails();
}

/** Store the time at the beginning of the planning process
 * @param argTime Set start_time to argTime
 */
inline void PlanStockManager::setStartTime(boost::posix_time::ptime const& argTime) {
	start_time = argTime;
}

/** Return the number of plans stored (best plans only)
 * @return The size of plans
 */
inline unsigned int PlanStockManager::getNbPlans() const {
	return plans.size();
}

#endif
