/*
 * Copyright (C) 2007 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include<iostream>

#include "SocialRuleBase.hh"
#include "WorldBase.hh"

using namespace std;

/** Copy constructor.
 * @param srb Base to copy
 */
SocialRuleBase::SocialRuleBase(SocialRuleBase const& srb)
{
	undesirableStates = srb.getUndesirableStates();
	undesirableSequencesFunctions = srb.getUndesirableSequencesFunctions();
	undesirableSequencesData = srb.getUndesirableSequencesData();
	effortBalancingRule = srb.getEffortBalancingRule();
	intricacyControlFunctions = srb.getIntricacyControlFunctions();
	intricacyControlData = srb.getIntricacyControlData();
	wastedTimeFunctions = srb.getWastedTimeFunctions();
	wastedTimeData = srb.getWastedTimeData();
	abstractionLegibilityRule = srb.getAbstractionLegibilityRule();
	badDecompositionsFunctions = srb.getBadDecompositionsFunctions();
	badDecompositionsData = srb.getBadDecompositionsData();
}

/** Add an undesirable state function to the base
 * @param name Name of the undesirable state rule to add
 * @param usf Pointer to the undesirable state function
 */
void SocialRuleBase::add(std::string const& name, undesirableStateFunc usf)
{
	if(undesirableStates.find(name) != undesirableStates.end())
		cerr << "WARNING: trying to add an undesirable state called \"" << name << "\" in the SocialRuleBase, name already used" << endl;
	else
		undesirableStates[name] = usf;
}

/** Add an undesirable sequence function to the base
 * @param name Name of the undesirable sequence rule to add
 * @param usfInit Pointer to the function to initialise the rule
 * @param usf Pointer to the undesirable sequence function
 */
void SocialRuleBase::add(std::string const& name, undesirableSequenceInitFunc usfInit, undesirableSequenceFunc usf)
{
	if(undesirableSequencesFunctions.find(name) != undesirableSequencesFunctions.end())
		cerr << "WARNING: trying to add an undesirable sequence called \"" << name << "\" in the SocialRuleBase, name already used" << endl;
	else if(usfInit == NULL)
		cerr << "ERROR: trying to add an undesirable sequence called \"" << name << "\" in the SocialRuleBase, cannot have an empty init function" << endl;
	else
		undesirableSequencesFunctions[name] = make_pair(usfInit, usf);
}

/** Add a bad decomposition function to the base.
 * @param name Name of the bad decomposition rule to add
 * @param bdfInit Pointer to the function to initialise the rule
 * @param bdf Pointer to the bad decomposition function
 */
void SocialRuleBase::add(std::string const& name, badDecompositionInitFunc bdfInit, badDecompositionFunc bdf)
{
	if(badDecompositionsFunctions.find(name) != badDecompositionsFunctions.end())
		cerr << "WARNING: trying to add a bad decomposition called \"" << name << "\" in the SocialRuleBase, name already used" << endl;
	else if(bdfInit == NULL)
		cerr << "ERROR: trying to add a bad decompositions called \"" << name << "\" in the SocialRuleBase, cannot have an empty init function" << endl;
	else
		badDecompositionsFunctions[name] = make_pair(bdfInit, bdf);
}

/** Call the init functions of social rules to get their data and store them.
 * Starts by freeing the memory, then call the functions to fill the datas (if the functions are defined)
 */
void SocialRuleBase::init()
{
	// we free memory
	map<string, pair<TaskList*, vector<string> > >::iterator it;
	for(it = undesirableSequencesData.begin() ; it != undesirableSequencesData.end() ; ++it)
		delete (it->second).first;

	map<string, badDecompoDescription>::iterator it2;
	unsigned int i;
	for(it2 = badDecompositionsData.begin() ; it2 != badDecompositionsData.end() ; ++it2)
		for(i=0 ; i<(it2->second).first.size() ; ++i)
			delete ((it2->second).first)[i];

	undesirableSequencesData.clear();
	badDecompositionsData.clear();

	map<string, pair<undesirableSequenceInitFunc, undesirableSequenceFunc> >::iterator it3;
	for(it3 = undesirableSequencesFunctions.begin() ; it3 != undesirableSequencesFunctions.end() ; ++it3)
		undesirableSequencesData[it3->first] = (it3->second).first();

	map<string, pair<badDecompositionInitFunc, badDecompositionFunc> >::iterator it4;
	for(it4 = badDecompositionsFunctions.begin() ; it4 != badDecompositionsFunctions.end() ; ++it4)
		badDecompositionsData[it4->first] = (it4->second).first();

	if(this->intricacyControlRuleDefined())
		intricacyControlData = intricacyControlFunctions.first();

	if(this->wastedTimeRuleDefined())
		wastedTimeData = wastedTimeFunctions.first();
}

/** Assignment operator.
 * @param other Base to assign this to.
 * @return The pointer to this
 */
SocialRuleBase& SocialRuleBase::operator=(SocialRuleBase const& other)
{
	undesirableStates = other.getUndesirableStates();
	undesirableSequencesFunctions = other.getUndesirableSequencesFunctions();
	undesirableSequencesData = other.getUndesirableSequencesData();
	effortBalancingRule = other.getEffortBalancingRule();
	intricacyControlFunctions = other.getIntricacyControlFunctions();
	intricacyControlData = other.getIntricacyControlData();
	wastedTimeFunctions = other.getWastedTimeFunctions();
	wastedTimeData = other.getWastedTimeData();
	abstractionLegibilityRule = other.getAbstractionLegibilityRule();
	badDecompositionsFunctions = other.getBadDecompositionsFunctions();
	badDecompositionsData = other.getBadDecompositionsData();

	this->init();
	return *this;
}
