/*
 * Copyright (C) 2007 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#if !defined HATP_WORLDBASE_REPORT_HH_
#define HATP_WORLDBASE_REPORT_HH_

#include<string>
#include<vector>
#include<typeinfo>
#include<boost/regex.hpp>
#include<boost/lexical_cast.hpp>

#include "Entity.hh"

#define VECTOR_REMOVE "vector_remove"
#define VECTOR_ADD "vector_add"
#define VECTOR_TEST_IN "vector_test_in"
#define VECTOR_TEST_NOT_IN "vector_test_not_in"
#define ATOM_CHANGE "atom_change"
#define ATOM_TEST_EQUAL "atom_test_equal"
#define ATOM_TEST_DIFF "atom_test_diff"

static const std::string REPORT_space = "([[:space:]]+)";
static const std::string REPORT_term = "([[:word:]]+)";
static const std::string REPORT_emptyterm = "EMPTY";

std::vector<std::string> parseModificationReport(const std::string& report);
std::vector<std::string> parseTestReport(const std::string& report);

/** The reports represent the preconditions and effects of actions.
 * A report is stored in a string in the form "<entity> <attribute> <operator> <value>".
 * Their are used to determine the causal links between actions see TimeProjection
 */
class WorldBaseReport
{
private:
	/** The vector containing all test reports (preconditions)
	*/
	std::vector<std::string> tests;

	/** The vector containing all modification reports (effects)
	*/
	std::vector<std::string> modifications;

public:
	WorldBaseReport();

	template<typename T> bool addAtomTestEqual(std::string const& entity, std::string const& attribute, T val);
	template<typename T> bool addAtomTestEqual(std::string const& entity, const char* attribute, T val);
	template<typename T> bool addAtomTestDiff(std::string const& entity, std::string const& attribute, T val);
	template<typename T> bool addAtomTestDiff(std::string const& entity, const char* attribute, T val);
	template<typename T> bool addVectorTestIn(std::string const& entity, std::string const& attribute, T val);
	template<typename T> bool addVectorTestIn(std::string const& entity, const char* attribute, T val);
	template<typename T> bool addVectorTestNotIn(std::string const& entity, std::string const& attribute, T val);
	template<typename T> bool addVectorTestNotIn(std::string const& entity, const char* attribute, T val);

	template<typename T> void addAtomModifChange(std::string const& entity, std::string const& attribute, T oldVal, T newVal);
	template<typename T> void addAtomModifChange(std::string const& entity, const char* attribute, T oldVal, T newVal);
	template<typename T> void addVectorModifAdd(std::string const& entity, std::string const& attribute, T val);
	template<typename T> void addVectorModifAdd(std::string const& entity, const char* attribute, T val);
	template<typename T> void addVectorModifRemove(std::string const& entity, std::string const& attribute, T val);
	template<typename T> void addVectorModifRemove(std::string const& entity, const char* attribute, T val);

	std::vector<std::string> const& getTestsReport() const;
	std::vector<std::string> const& getModificationsReport() const;

	bool clearTests();
	bool clearModifications();
	bool clearAll();
	~WorldBaseReport();
};

/** Constructor: empty
 */
inline WorldBaseReport::WorldBaseReport()
{}

/** Add an "equal" test on an atomic value.
 * @param entity Entity tested
 * @param attribute Entity attribute tested
 * @param val Value to test against
 * @return True
 */
template<typename T> bool WorldBaseReport::addAtomTestEqual(std::string const& entity, std::string const& attribute, T val)
{
	std::string value;
	try{ value = boost::lexical_cast<std::string>(val); }
	catch(...) {
		std::cerr << "ERROR when making a report from an equal test , type ";
		std::cerr << typeid(T).name() << " not convertible to a string" << std::endl;
	}
	std::string s = entity + " " + attribute + " ";
	s += ATOM_TEST_EQUAL;
	s += " ";
	if(value.empty()) s += REPORT_emptyterm;
	else s += value;
	tests.push_back(s);
	return true;
}

/** Add an "equal" test on an atomic value.
 * @param entity Entity tested
 * @param attribute Entity attribute tested
 * @param val Value to test against
 * @return True
 */
template<typename T> bool WorldBaseReport::addAtomTestEqual(std::string const& entity, const char* attribute, T val)
{
	std::string value;
	try{ value = boost::lexical_cast<std::string>(val); }
	catch(...) {
		std::cerr << "ERROR when making a report from an equal test , type ";
		std::cerr << typeid(T).name() << " not convertible to a string" << std::endl;
	}
	std::string s = entity + " " + std::string(attribute) + " ";
	s += ATOM_TEST_EQUAL;
	s += " ";
	if(value.empty()) s += REPORT_emptyterm;
	else s += value;
	tests.push_back(s);
	return true;
}

/** Add a "different" test on an atomic value.
 * @param entity Entity tested
 * @param attribute Entity attribute tested
 * @param val Value to test against
 * @return True
 */
template<typename T> bool WorldBaseReport::addAtomTestDiff(std::string const& entity, std::string const& attribute, T val)
{
	std::string value;
	try{ value = boost::lexical_cast<std::string>(val); }
	catch(...) {
		std::cerr << "ERROR when making a report from an equal test , type ";
		std::cerr << typeid(T).name() << " not convertible to a string" << std::endl;
	}
	std::string s = entity + " " + attribute + " ";
	s += ATOM_TEST_DIFF;
	s += " ";
	if(value.empty()) s += REPORT_emptyterm;
	else s += value;
	tests.push_back(s);
	return true;
}

/** Add a "different" test on an atomic value.
 * @param entity Entity tested
 * @param attribute Entity attribute tested
 * @param val Value to test against
 * @return True
 */
template<typename T> bool WorldBaseReport::addAtomTestDiff(std::string const& entity, const char* attribute, T val)
{
	std::string value;
	try{ value = boost::lexical_cast<std::string>(val); }
	catch(...) {
		std::cerr << "ERROR when making a report from an equal test , type ";
		std::cerr << typeid(T).name() << " not convertible to a string" << std::endl;
	}
	std::string s = entity + " " + std::string(attribute) + " ";
	s += ATOM_TEST_DIFF;
	s += " ";
	if(value.empty()) s += REPORT_emptyterm;
	else s += value;
	tests.push_back(s);
	return true;
}

/** Add a "is in" test on a set.
 * @param entity Entity tested
 * @param attribute Entity attribute tested
 * @param val Value to test against
 * @return True
 */
template<typename T> bool WorldBaseReport::addVectorTestIn(std::string const& entity, std::string const& attribute, T val)
{
	std::string value;
	try{ value = boost::lexical_cast<std::string>(val); }
	catch(...) {
		std::cerr << "ERROR when making a report from an equal test , type ";
		std::cerr << typeid(T).name() << " not convertible to a string" << std::endl;
	}
	std::string s = entity + " " + attribute + " ";
	s += VECTOR_TEST_IN;
	s += " ";
	if(value.empty()) s += REPORT_emptyterm;
	else s += value;
	tests.push_back(s);
	return true;
}

/** Add a "is in" test on a set.
 * @param entity Entity tested
 * @param attribute Entity attribute tested
 * @param val Value to test against
 * @return True
 */
template<typename T> bool WorldBaseReport::addVectorTestIn(std::string const& entity, const char* attribute, T val)
{
	std::string value;
	try{ value = boost::lexical_cast<std::string>(val); }
	catch(...) {
		std::cerr << "ERROR when making a report from an equal test , type ";
		std::cerr << typeid(T).name() << " not convertible to a string" << std::endl;
	}
	std::string s = entity + " " + std::string(attribute) + " ";
	s += VECTOR_TEST_IN;
	s += " ";
	if(value.empty()) s += REPORT_emptyterm;
	else s += value;
	tests.push_back(s);
	return true;
}

/** Add a "is not in" test on a set.
 * @param entity Entity tested
 * @param attribute Entity attribute tested
 * @param val Value to test against
 * @return True
 */
template<typename T> bool WorldBaseReport::addVectorTestNotIn(std::string const& entity, std::string const& attribute, T val)
{
	std::string value;
	try{ value = boost::lexical_cast<std::string>(val); }
	catch(...) {
		std::cerr << "ERROR when making a report from an equal test , type ";
		std::cerr << typeid(T).name() << " not convertible to a string" << std::endl;
	}
	std::string s = entity + " " + attribute + " ";
	s += VECTOR_TEST_NOT_IN;
	s += " ";
	if(value.empty()) s += REPORT_emptyterm;
	else s += value;
	tests.push_back(s);
	return true;
}

/** Add a "is not in" test on a set.
 * @param entity Entity tested
 * @param attribute Entity attribute tested
 * @param val Value to test against
 * @return True
 */
template<typename T> bool WorldBaseReport::addVectorTestNotIn(std::string const& entity, const char* attribute, T val)
{
	std::string value;
	try{ value = boost::lexical_cast<std::string>(val); }
	catch(...) {
		std::cerr << "ERROR when making a report from an equal test , type ";
		std::cerr << typeid(T).name() << " not convertible to a string" << std::endl;
	}
	std::string s = entity + " " + std::string(attribute) + " ";
	s += VECTOR_TEST_NOT_IN;
	s += " ";
	if(value.empty()) s += REPORT_emptyterm;
	else s += value;
	tests.push_back(s);
	return true;
}

/** Add a modification on an atom.
 * @param entity Entity changed
 * @param attribute Entity attribute changed
 * @param oldVal Previous value of the attribute
 * @param newVal New value for the attribute
 */
template<typename T>
void WorldBaseReport::addAtomModifChange(std::string const& entity, std::string const& attribute, T oldVal, T newVal)
{
	std::string oldValue, newValue;
	try{
			 oldValue = boost::lexical_cast<std::string>(oldVal);
			 newValue = boost::lexical_cast<std::string>(newVal);
		 }
	catch(...) {
		std::cerr << "ERROR when making a report from an atom modification , type ";
		std::cerr << typeid(T).name() << " not convertible to a string" << std::endl;
	}
	std::string s = entity + " " + attribute + " ";
	s += ATOM_CHANGE;
	s += " ";
	if(oldValue.empty()) s += REPORT_emptyterm;
	else s += oldValue;
	s += " ";
	if(newValue.empty()) s += REPORT_emptyterm;
	else s += newValue;
	modifications.push_back(s);
}

/** Add a modification on an atom.
 * @param entity Entity changed
 * @param attribute Entity attribute changed
 * @param oldVal Previous value of the attribute
 * @param newVal New value for the attribute
 */
template<typename T> void WorldBaseReport::addAtomModifChange(std::string const& entity, const char* attribute, T oldVal, T newVal)
{
	std::string oldValue, newValue;

	try{
			 oldValue = boost::lexical_cast<std::string>(oldVal);
			 newValue = boost::lexical_cast<std::string>(newVal);
		 }
	catch(...) {
		std::cerr << "ERROR when making a report from an atom modification , type ";
		std::cerr << typeid(T).name() << " not convertible to a string" << std::endl;
	}
	std::string s = entity + " " + std::string(attribute) + " ";
	s += ATOM_CHANGE;
	s += " ";
	if(oldValue.empty()) s += REPORT_emptyterm;
	else s += oldValue;
	s += " ";
	if(newValue.empty()) s += REPORT_emptyterm;
	else s += newValue;
	modifications.push_back(s);
}

/** Add a "add in" modification on a vector.
 * @param entity Entity changed
 * @param attribute Entity attribute changed
 * @param val Value to add in the attribute
 */
template<typename T> void WorldBaseReport::addVectorModifAdd(std::string const& entity, std::string const& attribute, T val)
{
	std::string value;
	try{ value = boost::lexical_cast<std::string>(val); }
	catch(...) {
		std::cerr << "ERROR when making a report from an add in set modifcation , type ";
		std::cerr << typeid(T).name() << " not convertible to a string" << std::endl;
	}
	std::string s = entity + " " + attribute + " ";
	s += VECTOR_ADD;
	s += " ";
	if(value.empty()) s += REPORT_emptyterm;
	else s += value;
	modifications.push_back(s);
}

/** Add a "add in" modification on a vector.
 * @param entity Entity changed
 * @param attribute Entity attribute changed
 * @param val Value to add in the attribute
 */
template<typename T> void WorldBaseReport::addVectorModifAdd(std::string const& entity, const char* attribute, T val)
{
	std::string value;
	try{ value = boost::lexical_cast<std::string>(val); }
	catch(...) {
		std::cerr << "ERROR when making a report from an add in set modifcation , type ";
		std::cerr << typeid(T).name() << " not convertible to a string" << std::endl;
	}
	std::string s = entity + " " + std::string(attribute) + " ";
	s += VECTOR_ADD;
	s += " ";
	if(value.empty()) s += REPORT_emptyterm;
	else s += value;
	modifications.push_back(s);
}

/** Add a "remove from" modification on a vector.
 * @param entity Entity changed
 * @param attribute Entity attribute changed
 * @param val Value to remove from the attribute
 */
template<typename T> void WorldBaseReport::addVectorModifRemove(std::string const& entity, std::string const& attribute, T val)
{
	std::string value;
	try{ value = boost::lexical_cast<std::string>(val); }
	catch(...) {
		std::cerr << "ERROR when making a report from an atom modifcation , type ";
		std::cerr << typeid(T).name() << " not convertible to a string" << std::endl;
	}
	std::string s = entity + " " + attribute + " ";
	s += VECTOR_REMOVE;
	s += " ";
	if(value.empty()) s += REPORT_emptyterm;
	else s += value;
	modifications.push_back(s);
}

/** Add a "remove from" modification on a vector.
 * @param entity Entity changed
 * @param attribute Entity attribute changed
 * @param val Value to remove from the attribute
 */
template<typename T> void WorldBaseReport::addVectorModifRemove(std::string const& entity, const char* attribute, T val)
{
	std::string value;
	try{ value = boost::lexical_cast<std::string>(val); }
	catch(...) {
		std::cerr << "ERROR when making a report from an atom modifcation , type ";
		std::cerr << typeid(T).name() << " not convertible to a string" << std::endl;
	}
	std::string s = entity + " " + std::string(attribute) + " ";
	s += VECTOR_REMOVE;
	s += " ";
	if(value.empty()) s += REPORT_emptyterm;
	else s += value;
	modifications.push_back(s);
}


/** Return all test reports.
 * @return tests
 */
inline std::vector<std::string> const& WorldBaseReport::getTestsReport() const
{
	return tests;
}

/** Return all modification reports/
 * @return modifications
 */
inline std::vector<std::string> const& WorldBaseReport::getModificationsReport() const
{
	return modifications;
}

/** Empty the test reports.
 * @return True
 */
inline bool WorldBaseReport::clearTests()
{
	tests.clear();
	return true;
}

/** Empty the modification reports.
 * @return True
 */
inline bool WorldBaseReport::clearModifications()
{
	modifications.clear();
	return true;
}

/** Empty both test and modification reports.
 * @return true
 */
inline bool WorldBaseReport::clearAll()
{
	tests.clear();
	modifications.clear();
	return true;
}

/** Destructor: clear tests and modifications
 */
inline WorldBaseReport::~WorldBaseReport()
{
	tests.clear();
	modifications.clear();
}

#endif
